# This file is based on code from:
# - https://github.com/pytorch/pytorch/blob/v2.0.1/torch/optim/swa_utils.py
# - https://github.com/pytorch/pytorch/blob/main/torch/optim/swa_utils.py
# It implements Stochastic Weight Averaging-Gaussian (SWAG) from the paper
# "A Simple Baseline for Bayesian Uncertainty in Deep Learning" by Wesley Maddox et al.
# Some functions are also adapted from https://github.com/wjmaddox/swa_gaussian/tree/master.

# Author: J. Hu

# Relevant license content follows.

# From PyTorch:

# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# From Caffe2:

# Copyright (c) 2016-present, Facebook Inc. All rights reserved.

# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.

# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.

# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.

# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain

# All contributions by Cruise LLC:
# Copyright (c) 2022 Cruise LLC.
# All rights reserved.

# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.

# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.

# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
#    and IDIAP Research Institute nor the names of its contributors may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# BSD 2-Clause License

# Copyright (c) 2019, Wesley Maddox, Timur Garipov, Pavel Izmailov,  Dmitry Vetrov, Andrew Gordon Wilson
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
from copy import deepcopy
from typing import Union

import torch
from torch import Tensor
from torch.nn import Module


@torch.no_grad()
def swa_update(
    averaged_param: Tensor, current_param: Tensor, num_averaged: Union[Tensor, int]
):
    return averaged_param + (current_param - averaged_param) / (num_averaged + 1)


@torch.no_grad()
def second_moment_update(
    averaged_second_moment: Tensor,
    current_param: Tensor,
    num_averaged: Union[Tensor, int],
):
    return averaged_second_moment + (
        current_param * current_param - averaged_second_moment
    ) / (num_averaged + 1)


def flatten(lst):
    tmp = [i.contiguous().view(-1, 1) for i in lst]
    return torch.cat(tmp).view(-1)


def unflatten_like(vector, likeTensorList):
    # Takes a flat torch.tensor and unflattens it to a list of torch.tensors
    #    shaped like likeTensorList
    outList = []
    i = 0
    for tensor in likeTensorList:
        n = tensor.numel()
        outList.append(vector[:, i : i + n].view(tensor.shape))
        i += n
    return outList


class SwagModel(Module):
    r"""Implements averaged model for Stochastic Weight Averaging Gaussian (SWAG).

    SWAG was proposed in `A Simple Baseline for Bayesian Uncertainty
    in Deep Learning` by Wesley J. Maddox, Timur Garipov, Pavel Izmailov,
    Dmitry Vetrov and Andrew Gordon Wilson in 2019.

    SwagModel is a modified version of `torch.optim.swa_utils.AveragedModel` (`pytorch` v2.0.1), which
    implements Stochastic Weight Averaging (SWA). This modification is inspired by
    the original implementation of SWAG available at: https://github.com/wjmaddox/swa_gaussian.

    SwagModel class creates a copy of the provided module :attr:`model`
    on the device :attr:`device` and allows to compute running averages of the
    parameters of the :attr:`model`.

    Args:
        model (torch.nn.Module): model to use with SWAG
        device (torch.device, optional): if provided, the averaged model will be
            stored on the :attr:`device` (default: ``None``)
        avg_fn (function): the averaging function used to update
            parameters; the function must take in the current value of the
            :class:`SwagModel` parameter, the current value of :attr:`model`
            parameter and the number of models already averaged; if None,
            equally weighted average is used (default: ``swa_update``)
        use_buffers (bool): if ``True``, it will compute running averages for
            both the parameters and the buffers of the model. (default: ``False``)

    .. note::
        :attr:`avg_fn` is not saved in the :meth:`state_dict` of the model.

    .. note::
        When :meth:`update_parameters` is called for the first time (i.e.
        :attr:`n_averaged` is `0`) the parameters of `model` are copied
        to the parameters of :class:`SwagModel`. For every subsequent
        call of :meth:`update_parameters` the function `avg_fn` is used
        to update the parameters.
    """

    def __init__(
        self, model: Module, device=None, avg_fn=swa_update, use_buffers=False
    ):
        super().__init__()
        self.module = deepcopy(model)
        if device is not None:
            self.module = self.module.to(device)
        self.register_buffer(
            "n_averaged", torch.tensor(0, dtype=torch.long, device=device)
        )
        self.avg_fn = avg_fn
        self.use_buffers = use_buffers
        self._init_swag_params()

    def _init_swag_params(self):
        for i, p in enumerate(self.module.parameters()):
            sec_moment_name = f"sec_moment_{i}"
            sec_moment_tensor = torch.zeros_like(p)
            self.register_buffer(sec_moment_name, sec_moment_tensor)
            dev_name = f"dev_{i}"
            dev_tensor = p.new_empty((p.numel(), 0)).zero_()
            self.register_buffer(dev_name, dev_tensor)

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def update_parameters(self, model: Module):
        self_param = (
            itertools.chain(self.module.parameters(), self.module.buffers())
            if self.use_buffers
            else self.parameters()
        )

        model_param = (
            itertools.chain(model.parameters(), model.buffers())
            if self.use_buffers
            else model.parameters()
        )

        for i, (p_swa, p_model) in enumerate(zip(self_param, model_param)):
            device = p_swa.device
            p_swa_ = p_swa.detach()
            p_model_ = p_model.detach().to(device)
            if self.n_averaged == 0:
                p_swa_.copy_(p_model_)
                setattr(self, f"sec_moment_{i}", p_model_ * p_model_)

            else:
                p_swa_.copy_(self.avg_fn(p_swa_, p_model_, self.n_averaged.to(device)))

                sec_moment_value = second_moment_update(
                    getattr(self, f"sec_moment_{i}"),
                    p_model_,
                    self.n_averaged.to(device),
                )
                setattr(self, f"sec_moment_{i}", sec_moment_value)

                dev = (p_model_ - p_swa_).view(-1, 1)
                dev_value = torch.cat((getattr(self, f"dev_{i}"), dev), dim=1)

                setattr(self, f"dev_{i}", dev_value)

        if not self.use_buffers:
            # If not apply running averages to the buffers,
            # keep the buffers in sync with the source model.
            for b_swa, b_model in zip(self.module.buffers(), model.buffers()):
                b_swa.detach().copy_(b_model.detach().to(device))

        self.n_averaged += 1

    def sample(self, scale=0.5, cov=True, var_clamp=1e-30):
        """Draw SWAG sample model.

        Args:
             scale (float): The scale parameter for covariance. Default: ``0.5``. When scale = 0.0, the ``sample`` function returns a SWA model.
             cov (bool): If True, enable calculating low-rank covariance. ``True``.
             var_clamp (float): A small positive value used to clamp the computed variance to prevent numerical issues. Default: ``1.e-30``

         Returns:
             sample_model (rnn model): A sampled SWAG model
        """
        sample_model = deepcopy(self.module)

        scale_sqrt = scale**0.5

        # get the mean, second moment
        mean_list = [p for p in self.module.parameters()]
        mean = flatten(mean_list)
        if torch.isnan(mean).any():
            print("mean contains NAN value")
            breakpoint()
        num_param = len(list(self.module.parameters()))
        sec_moment_list = [getattr(self, f"sec_moment_{i}") for i in range(num_param)]
        sec_moment = flatten(sec_moment_list)

        # draw diagonal variance sample
        var = torch.clamp(sec_moment - mean**2, var_clamp)
        with torch.no_grad():
            var_sample = var.sqrt() * torch.randn_like(var)
        # temp: just for debugging
        self.var_sample = var_sample

        if torch.isnan(var_sample).any():
            print("var_sample contain NAN value")
            breakpoint()

        # if covariance draw low rank sample
        if cov:
            dev_list = [getattr(self, f"dev_{i}") for i in range(num_param)]
            dev = torch.cat(dev_list, dim=0)
            cov_sample = dev.matmul(
                dev.new_empty((dev.size(1),), requires_grad=False).normal_()
            )

            # we start deviation collect when self.n_averaged = 1 instead of 0
            # so K = self.n_averaged -1
            cov_sample /= (self.n_averaged - 2) ** 0.5

            if torch.isnan(cov_sample).any():
                print("cov_sample contains NAN value")
                breakpoint()
            rand_sample = var_sample + cov_sample
        else:
            rand_sample = var_sample

        # update sample with mean and scale
        sample = mean + scale_sqrt * rand_sample
        sample = sample.unsqueeze(0)

        # unflatten new sample like the mean sample
        samples_list = unflatten_like(sample, mean_list)

        # update sample model
        for p_sample, sample_tensor in zip(sample_model.parameters(), samples_list):
            p_sample_ = p_sample.detach()
            p_sample_.copy_(sample_tensor)

        return sample_model



import re
import os
import itertools
import numpy as np
import pandas as pd
import seaborn as sns
from pathlib import Path
import matplotlib.pylab as plt
from matplotlib import gridspec
from matplotlib.lines import Line2D
from matplotlib.ticker import FixedLocator
from astropy.cosmology import FlatLambdaCDM
from sklearn.metrics import confusion_matrix

from ..utils import data_utils as du
from ..utils import logging_utils as lu
from ..utils import performance_utils as pu
from ..utils import visualization_utils as vu

# Plotting styles
from ..utils.visualization_utils import (
    ALL_COLORS,
    BI_COLORS,
    CONTRAST_COLORS,
    MARKER_DIC,
    FILL_DIC,
    MARKER_LIST,
    CMAP,
    get_model_visualization_name,
)

cosmo = FlatLambdaCDM(H0=70, Om0=0.295)

#################
# Plot utils
#################


def class_target_decode(target):
    if target == 2:
        return "Binary"
    if target == 3:
        return "Ternary"
    else:
        return "Seven-way"


def dist_mu(redshift):
    mu = cosmo.distmod(redshift)

    return mu.value


def create_sigma_df(df_grouped, class_=0):
    """From grouped prediction df create a df with sigma values"""
    sigma_all_list = []
    sigma_peak_list = []
    snid_list = []
    pred_class_list = []
    for SNID, SNID_df in df_grouped:
        arr_proba = SNID_df[f"all_class{class_}"]
        perc_16 = np.percentile(arr_proba, 16)
        perc_84 = np.percentile(arr_proba, 84)
        sigma_all_list.append(perc_84 - perc_16)

        arr_proba = SNID_df[f"PEAKMJD_class{class_}"]
        perc_16 = np.percentile(arr_proba, 16)
        perc_84 = np.percentile(arr_proba, 84)
        sigma_peak_list.append(perc_84 - perc_16)
        snid_list.append(SNID)

        # get predicition for this SNID
        k_all_probas = [k for k in SNID_df.keys() if "all_class" in k]
        median_prob_forSNID = SNID_df[k_all_probas].median()
        pred_class = median_prob_forSNID.idxmax()
        arr_proba = SNID_df[pred_class]
        # get sigma for this class
        perc_16 = np.percentile(arr_proba, 16)
        perc_84 = np.percentile(arr_proba, 84)
        pred_class_list.append(perc_84 - perc_16)

    df = pd.DataFrame()
    df["SNID"] = np.array(snid_list)
    df["sigma_all"] = np.array(sigma_all_list)
    df["sigma_peak"] = np.array(sigma_peak_list)
    df["pred_sigma_all"] = np.array(pred_class_list)
    return df


def plot_acc_vs_nsn(df, settings):
    """Plot accuracy vs number of SNe used for training

    Args:
        df (DataFrame): prediction dataframe
        settings (ExperimentSettings): custom class to hold hyperparameters

    Returns:
        figure (png)
    """
    plt.clf()
    fig = plt.figure()
    fig, ax1 = plt.subplots()
    ax1.grid(True)
    ax1.set_axisbelow(True)

    models_list = ["randomforest", "vanilla"]
    redshift_list = df["redshift"].unique()

    label_dic = {"randomforest": "Random Forest", "vanilla": "Baseline RNN"}

    group_cols = ["model_name_noseed", "model_type", "redshift", "data_fraction"]
    keep_cols = group_cols + ["all_accuracy"]

    # Cast to float for groupby operation (all_accuracy is type `O`)
    df.all_accuracy = df.all_accuracy.astype(float)

    df_errorbars = (
        df[keep_cols]
        .groupby(group_cols)
        .mean()
        .rename(columns={"all_accuracy": "all_accuracy_mean"})
        .reset_index()
    )
    df_errorbars["all_accuracy_std"] = (
        df[keep_cols]
        .groupby(group_cols)
        .std()
        .rename(columns={"all_accuracy": "all_accuracy_std"})
        .reset_index()["all_accuracy_std"]
    )

    for i, basemodel in enumerate(models_list):
        for z in redshift_list:
            df_sel = df_errorbars[
                (df_errorbars["model_type"] == basemodel)
                & (df_errorbars["redshift"] == z)
            ]
            # Plot these independently to avoid polluting legend
            ax1.errorbar(
                df_sel["data_fraction"],
                df_sel["all_accuracy_mean"],
                yerr=df_sel["all_accuracy_std"],
                c=CONTRAST_COLORS[i],
                fmt="none",
                zorder=3 if basemodel == "vanilla" else 1,
            )
            ax1.plot(
                df_sel["data_fraction"],
                df_sel["all_accuracy_mean"],
                label=label_dic[basemodel],
                marker=MARKER_DIC[basemodel],
                c=CONTRAST_COLORS[i],
                fillstyle=FILL_DIC[z],
                lw=0,
                markersize=10,
                markeredgewidth=1.5,
            )
    legend_elements = [
        Line2D(
            [0],
            [0],
            marker="s",
            lw=0,
            color="indigo",
            label="Baseline RNN",
            markerfacecolor="w",
            markersize=12,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            lw=0,
            color="darkorange",
            label="Random Forest",
            markerfacecolor="w",
            markersize=12,
        ),
    ]

    ax1.legend(handles=legend_elements, loc=4)
    ax1.set_ylabel("accuracy", fontsize=18)
    ax1.set_ylim(91, 100)
    ax1.set_xlim(0.025)
    ax1.set_xlabel("# SNe for training", fontsize=18)

    # exchange axis and reformat
    ax2 = ax1.twiny()
    ax1Xs = [round(i, 1) for i in ax1.get_xticks()]
    ax2Xs = []
    for X in ax1Xs:
        # BEWARE: only valid with SALTfitted sample
        ax2Xs.append("{:0.1e}".format(int(X * 881_969 * 0.8)))

    ax1.set_xticklabels(ax2Xs)
    ax2.set_xticks(ax1Xs)
    ax2.set_xbound(ax1.get_xbound())
    ax2.set_xticklabels(ax1Xs)

    title = ax1.set_title("data fraction", fontsize=18)
    title.set_y(1.1)
    plt.tight_layout()

    fig.subplots_adjust(top=0.85)
    fig.savefig(f"{settings.figures_dir}/accuracy_vs_nSN.png")
    plt.close()
    plt.clf()


def plot_calibration(settings, prediction_files=None):
    """Plot reliability diagram

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters

    Returns:
        figure (png)
    """

    if settings.prediction_files is None and prediction_files is None:

        lu.print_yellow(
            "Warning: Calibration plot cannot be generated. Please provide at least one prediction file."
        )

        return

    elif prediction_files is not None:
        if isinstance(prediction_files, str):
            prediction_files = [prediction_files]

        metric_files = [
            pf.replace("PRED", "METRICS")
            for pf in prediction_files
            if os.path.exists(pf.replace("PRED", "METRICS"))
        ]
        tmp_not_found = [
            pf.replace("PRED", "METRICS")
            for pf in prediction_files
            if not os.path.exists(pf.replace("PRED", "METRICS"))
        ]

        if len(tmp_not_found) > 0:
            lu.print_red(f"Files not found {tmp_not_found}")

        if len(metric_files) == 0:
            lu.print_yellow(
                "Warning: Metric file is not found. Calibration plot cannot be generated."
            )
            return

    else:
        metric_files = [
            f.replace("PRED", "METRICS")
            for f in settings.prediction_files
            if os.path.exists(f.replace("PRED", "METRICS"))
        ]
        tmp_not_found = [
            f.replace("PRED", "METRICS")
            for f in settings.prediction_files
            if not os.path.exists(f.replace("PRED", "METRICS"))
        ]

        if len(tmp_not_found) > 0:
            lu.print_red(f"Files not found {tmp_not_found}")

        if len(metric_files) == 0:
            lu.print_yellow(
                "Warning: Metric file is not found. Calibration plot cannot be generated."
            )
            return

    plt.figure(figsize=(10, 10))
    ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
    ax11 = plt.subplot2grid((3, 1), (2, 0), sharex=ax1)

    ax1.plot([0, 1], [0, 1], "k:", label="Perfectly calibrated")

    plot_path = f"{settings.figures_dir}/calibration"
    Path(plot_path).mkdir(parents=True, exist_ok=True)

    for i, f in enumerate(metric_files):
        df = pd.read_pickle(f)

        mean_bins, TPF = df["calibration_mean_bins"][0], df["calibration_TPF"][0]
        model_name = df["model_name"][0]

        # add swag information to the label
        swag = ""
        if f.endswith("_swag.pickle"):
            swag = "_swag"
        if f.endswith("_swa.pickle"):
            swag = "_swa"

        calibration_dispersion = TPF - mean_bins
        ax1.plot(
            mean_bins,
            TPF,
            "s-",
            color=ALL_COLORS[i],
            label=get_model_visualization_name(df["model_name"][0]) + swag,
            marker=MARKER_LIST[i],
        )
        ax11.scatter(
            mean_bins,
            calibration_dispersion,
            color=ALL_COLORS[i],
            marker=MARKER_LIST[i],
        )

    ax1.set_ylabel("fraction of positives (fP)", fontsize=18)
    ax1.set_ylim([-0.05, 1.05])
    ax11.set_xlabel("mean predicted probability", fontsize=18)
    ax1.legend()
    plt.setp(ax1.get_xticklabels(), visible=False)

    ax11.set_ylabel("residual fP", fontsize=18)
    ax11.set_ylim([-0.2, 0.2])
    ax11.plot([0, 1], np.zeros(len([0, 1])), "k:")
    ax11.plot([0, 1], 0.1 * np.ones(len([0, 1])), ":", color="grey")
    ax11.plot([0, 1], -0.1 * np.ones(len([0, 1])), ":", color="grey")
    plt.setp(ax11.get_xticklabels(), visible=False)

    if len(metric_files) == 1:
        nameout = f"{plot_path}/calib_{model_name}.png"
    else:
        nameout = f"{plot_path}/calib_multimodels.png"

    plt.savefig(nameout)
    plt.close()
    plt.clf()
    lu.print_green("Finished generating calibration plots.")


def plot_confusion_matrix(
    settings, cm, classes, normalize=False, cmap=vu.CMAP, nameout=None
):
    """Plot confusion matrix
    Based on sklearn tutorial

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        cm (np.array): confusion matrix
        classes (list): name of classes in cm
        normalize (optional): Normalization can be applied by setting `normalize=True`
        cmap (colormap)
        nameout (optional): out name for figure

    Returns:
        figure (png)
    """
    if normalize:
        cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]

    plt.imshow(cm, interpolation="nearest", cmap=cmap)
    plt.colorbar()
    # cbar.set_clim(0, 1.0)
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = ".2f" if normalize else "d"
    thresh = cm.max() / 2.0
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(
            j,
            i,
            format(cm[i, j], fmt),
            horizontalalignment="center",
            color="white" if cm[i, j] > thresh else "black",
        )

    plt.ylabel("True label", fontsize=18)
    plt.xlabel("Predicted label", fontsize=18)
    plt.yticks(fontsize=14)
    plt.xticks(fontsize=14)
    plot_path = f"{settings.figures_dir}/cnf_matrix"
    os.makedirs(plot_path, exist_ok=True)
    plt.tight_layout()
    plt.savefig(f"{plot_path}/{nameout}.png")


def multiplot_violin_paper(df, fname, settings):
    """Plot data properties as violin plots.

    Far from optimized code: seaborn does not make this easy so added
    a lot of formatting using raw matplotlib commands

    Args:
        df (DataFrame): prediction dataframe
        fname (filename):
        settings (ExperimentSettings): custom class to hold hyperparameters

    Returns:
        figure (png)
    """

    # Set up the axes with gridspec
    plt.clf()
    fig = plt.figure()
    grid = plt.GridSpec(2, 4, hspace=0.2, wspace=0.2)
    ax_00 = fig.add_subplot(grid[0, 0])
    ax_01 = fig.add_subplot(grid[0, 1], sharey=ax_00)
    ax_02 = fig.add_subplot(grid[0, 2], sharey=ax_00)
    ax_03 = fig.add_subplot(grid[0, 3], sharey=ax_00)
    ax_1 = fig.add_subplot(grid[1, :])

    axes = [ax_00, ax_01, ax_02, ax_03]

    # Ia vs non Ia
    sns.set_palette(sns.color_palette(BI_COLORS))
    g = sns.violinplot(
        x="target",
        y="SIM_PEAKMAG_g",
        hue="salt",
        data=df,
        split=True,
        ax=axes[0],
        inner="quartile",
    )
    g.set_xlabel("")
    g.legend_.remove()
    g.yaxis.set_tick_params(labelsize=14)
    g.set_title("g", fontsize=14)
    g.set_ylabel("magnitude", fontsize=14)
    g.set_ylim(20, 28)
    g.spines["right"].set_visible(False)
    g.spines["top"].set_visible(False)
    g.xaxis.set_major_locator(FixedLocator([0, 1]))
    g.set_xticklabels([getattr(settings, "target_sntype", "Ia"), f"non{getattr(settings, 'target_sntype', 'Ia')}"], fontsize=14)

    g = sns.violinplot(
        x="target",
        y="SIM_PEAKMAG_i",
        hue="salt",
        data=df,
        split=True,
        ax=axes[1],
        inner="quartile",
    )
    g.set_xlabel("")
    g.set_ylabel("")
    g.legend_.remove()
    g.yaxis.set_ticks_position("none")
    g.set_title("i", fontsize=14)
    g.set_ylim(20, 28)
    g.spines["right"].set_visible(False)
    g.spines["top"].set_visible(False)
    g.spines["left"].set_visible(False)
    g.xaxis.set_major_locator(FixedLocator([0, 1]))
    g.set_xticklabels([getattr(settings, "target_sntype", "Ia"), f"non{getattr(settings, 'target_sntype', 'Ia')}"], fontsize=14)
    plt.setp(axes[1].get_yticklabels(), visible=False)

    g = sns.violinplot(
        x="target",
        y="SIM_PEAKMAG_r",
        hue="salt",
        data=df,
        split=True,
        ax=axes[2],
        inner="quartile",
    )
    g.legend_.remove()
    g.yaxis.set_ticks_position("none")
    g.set_xlabel("")
    g.set_ylabel("")
    g.set_title("r", fontsize=18)
    g.xaxis.set_tick_params(labelsize=14)
    g.set_ylim(20, 28)
    g.spines["right"].set_visible(False)
    g.spines["top"].set_visible(False)
    g.spines["left"].set_visible(False)
    g.xaxis.set_major_locator(FixedLocator([0, 1]))
    g.set_xticklabels([getattr(settings, "target_sntype", "Ia"), f"non{getattr(settings, 'target_sntype', 'Ia')}"], fontsize=14)
    plt.setp(axes[2].get_yticklabels(), visible=False)

    g = sns.violinplot(
        x="target",
        y="SIM_PEAKMAG_z",
        hue="salt",
        data=df,
        split=True,
        ax=axes[3],
        inner="quartile",
    )
    g.legend_.remove()
    g.yaxis.set_ticks_position("none")
    g.set_title("z", fontsize=14)
    g.set_xlabel("")
    g.set_ylabel("")
    g.set_ylim(20, 28)
    g.spines["right"].set_visible(False)
    g.spines["top"].set_visible(False)
    g.spines["left"].set_visible(False)
    g.xaxis.set_major_locator(FixedLocator([0, 1]))
    g.set_xticklabels([getattr(settings, "target_sntype", "Ia"), f"non{getattr(settings, 'target_sntype', 'Ia')}"], fontsize=14)
    plt.setp(axes[3].get_yticklabels(), visible=False)

    # redshift
    g = sns.violinplot(
        x=settings.sntype_var,
        y="SIM_REDSHIFT_CMB",
        hue="salt",
        data=df,
        split=True,
        ax=ax_1,
        inner="quartile",
    )
    g.set_ylabel("simulated redshift", fontsize=14)
    g.set_xlabel("")
    g.set_ylim(0, 1.0)
    # Extract labels in consistent order (e.g., sorted by key)
    tick_vals = list(sorted(settings.sntypes.keys()))
    tick_labels = [settings.sntypes[k] for k in tick_vals]
    g.set_xticks(tick_vals)
    g.set_xticklabels(tick_labels, fontsize=14)
    g.xaxis.set_tick_params(labelsize=14)
    g.yaxis.set_tick_params(labelsize=14)
    g.legend_.remove()
    g.spines["right"].set_visible(False)
    g.spines["top"].set_visible(False)
    g.spines["bottom"].set_visible(False)

    plt.savefig(f"{settings.figures_dir}/multiviolin_{fname}.png")
    plt.close()
    del fig


def binned_2d(
    bin_centers,
    y_dic,
    xname,
    yname,
    label_list,
    color_sequence,
    MARKER_LIST,
    nameout,
    extra_line=None,
):
    """Plot scatter plot

    Args:
        bin_centers (np.array): x centers
        y_dic (dict): y values
        xname (str): xlabel
        yname (str): ylabel
        label_list (list(str)): for legend
        color_sequence (list(str)): color code
        MARKER_LIST (list(str)): marker code
        nameout (str): out name of figure (.png)
        extra_line (optional): extra line on 0 and 100 (y)

    Returns:
        figure (png)
    """
    plt.clf()
    fig = plt.figure()
    ax = plt.gca()
    for i, k in enumerate(y_dic.keys()):
        ax.scatter(
            bin_centers,
            y_dic[k],
            color=color_sequence[i],
            label=label_list[i],
            marker=MARKER_LIST[i],
        )
    if extra_line:
        ax.plot(
            ax.get_xlim(), np.zeros(len(ax.get_xlim())), color="grey", linestyle="--"
        )
        ax.plot(
            ax.get_xlim(),
            100 * np.ones(len(ax.get_xlim())),
            color="grey",
            linestyle="--",
        )
    ax.set_ylim(0.01, 1)
    ax.set_yscale("log")
    ax.set_xlabel(xname, fontsize=18)
    ax.set_ylabel(yname, fontsize=18)
    ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    plt.tight_layout()
    plt.savefig(nameout)
    plt.clf()
    plt.close()
    del fig


def plot_acc_matrix(acc_dic, n_meas_dic, settings, nameout):
    """Plot accuracy matrix

    Args:
        acc_dic (dict): with np.arrays of accuracy
        n_meas_dic (dict): with n measurements required for this accuracy
        settings (ExperimentSettings): custom class to hold hyperparameters
        nameout (str): outname for plot
    Returns:
        figure (png)
    """

    plt.clf()
    fig = plt.figure(figsize=(4, 6))
    ax = fig.add_subplot(111)

    min_length = min([len(acc_dic[band]) for band in settings.list_filters])
    x_labels = settings.list_filters
    y_labels = [int(round(n)) for n in n_meas_dic["g"]][:min_length]
    CMAP.set_bad(color="white")
    acc_mat = np.vstack([acc_dic[band][:min_length] for band in settings.list_filters])
    acc_mat = acc_mat.transpose()
    cax = ax.matshow(acc_mat, cmap=CMAP, vmin=50, vmax=100.0)
    fig.colorbar(cax)
    ax.tick_params(axis="both", labelsize=14)
    ax.set_xticks(np.arange(0, len(x_labels), 1))
    ax.set_xticklabels(x_labels)
    ax.set_yticks(np.arange(0, len(y_labels), 1))
    ax.set_yticklabels(y_labels)
    plt.tight_layout()
    plt.savefig(nameout)
    plt.close()
    del fig


def plot_HDres_histos_vs_z(
    df,
    nameout,
    threshold_var="class0",
    threshold_list=[0.5, 0.7, 0.9],
    threshold_sign=">",
    settings=None,
):
    """Plot Hubble diagram residuals and histograms
    selects class sample and performas HD

    Args:
        df (DataFrame): predictions
        nameout (str): outname figure
        threshold_var (str): which class used for sample selection
        threshold_list (list): list of probability threshold for selecting sample
        threshold_sign (str): sign of the probability threshold (e.g. ">" )
    Returns:
        figure (png)
    """

    P = df[df["class0"] > 0.5]
    Ias = df[df["target"] == 0]

    TP = P[P["target"] == 0]
    FP = P[P["target"] != 0]

    sel_TP_dic = {}
    sel_FP_dic = {}
    for t in threshold_list:
        if threshold_sign == ">":
            sel_TP_dic[t] = TP[TP[threshold_var] > t]
            sel_FP_dic[t] = FP[FP[threshold_var] > t]
        else:
            sel_TP_dic[t] = TP[TP[threshold_var] < t]
            sel_FP_dic[t] = FP[FP[threshold_var] < t]

    plt.clf()
    cm = CMAP
    fig = plt.figure(figsize=(14, 14))
    # gs = gridspec.GridSpec(4, 2, width_ratios=[3, 1], height_ratios=[2, 2, 1, 1])
    # gs.update(wspace=0.1, hspace=0.3)

    # # gridspec init
    # ax00 = plt.subplot(gs[0, 0])  # Hres Ia
    # ax10 = plt.subplot(gs[1, 0], sharex=ax00)  # Hres CC
    # ax20 = plt.subplot(gs[2:, 0], sharex=ax00)  # efficiency
    # ax01 = plt.subplot(gs[0, 1], sharey=ax00)  # histo Ia
    # ax11 = plt.subplot(gs[1, 1], sharey=ax10)  # histo CC
    # ax21 = plt.subplot(gs[2, 1])  # histo x1
    # ax31 = plt.subplot(gs[3, 1])  # histo c
    gs = gridspec.GridSpec(3, 3, height_ratios=[2, 2, 1])
    # gs.update(wspace=0.2, hspace=0.1)

    # gridspec init
    ax00 = plt.subplot(gs[0, 0:2])  # Hres Ia
    ax10 = plt.subplot(gs[1, 0:2], sharex=ax00)  # Hres CC
    ax20 = plt.subplot(gs[2, 0])  # redshift dist
    ax01 = plt.subplot(gs[0, 2], sharey=ax00)  # histo Ia
    ax11 = plt.subplot(gs[1, 2], sharey=ax10)  # histo CC
    ax21 = plt.subplot(gs[2, 1])  # histo x1
    ax31 = plt.subplot(gs[2, 2])  # histo c

    # lines
    ax00.plot([0, 1.2], np.zeros(len([0, 1.2])), "k:")
    ax10.plot([0, 1.2], np.zeros(len([0, 1.2])), "k:")

    mubins = np.arange(-2, 2 + 0.1, 0.1)

    # Hres w. histogram
    def HRwhisto(
        df, sel_dic, ax_left, ax_right, threshold_sign, ylabel="TP", visible=False
    ):
        target_class = getattr(settings, "target_sntype", "Ia") if settings else "Ia"
        if ylabel == "TP":
            sntyp = target_class
        else:
            sntyp = "CC"
        ax_left.scatter(
            df["SIM_REDSHIFT_CMB"],
            df["delmu"],
            c=df["class0"],
            cmap=CMAP,
            vmin=0.5,
            vmax=1,
            s=8,
        )
        ax_left.errorbar(
            df["SIM_REDSHIFT_CMB"],
            df["delmu"],
            yerr=df["delmu_err"],
            color="gray",
            zorder=0,
            fmt="none",
            marker="none",
        )

        ax_left.set_ylim(-2, 2)
        ax_left.set_xlim(0, 1.2)
        ax_left.set_ylabel(f"{ylabel} residual", fontsize=18)
        ax_left.tick_params(labelsize=14)
        plt.setp(ax_left.get_xticklabels(), visible=visible)
        if visible is True:
            ax_left.set_xlabel("simulated redshift", fontsize=18)
        for t in threshold_list:
            sel = sel_dic[t]
            n_SNe = len(sel)
            ax_right.hist(
                sel["delmu"],
                orientation="horizontal",
                histtype="step",
                color=cm(t),
                bins=mubins,
                density=True,
                label=f"{n_SNe} {sntyp} {threshold_sign} {t}",
                lw=2,
            )
        ax_right.legend(loc="lower center", prop={"size": 13})
        plt.setp(ax_right.get_yticklabels(), visible=False)
        plt.setp(ax_right.get_xticklabels(), visible=False)
        ax_right.plot(
            [ax_right.get_xlim()[0], ax_right.get_xlim()[1]],
            np.zeros(len([ax_right.get_xlim()[0], ax_right.get_xlim()[1]])),
            "k:",
        )

    HRwhisto(TP, sel_TP_dic, ax00, ax01, threshold_sign, ylabel="TP", visible=False)
    HRwhisto(FP, sel_FP_dic, ax10, ax11, threshold_sign, ylabel="FP", visible=True)

    # z histos
    n, bins_to_use, tmp = ax20.hist(
        Ias["SIM_REDSHIFT_CMB"], histtype="step", color="black", bins=15, lw=3
    )

    for t in threshold_list:
        sel_TP = sel_TP_dic[t]
        sel_FP = sel_FP_dic[t]
        ax20.hist(
            sel_TP["SIM_REDSHIFT_CMB"], histtype="step", color=cm(t), bins=bins_to_use
        )
        ax20.hist(
            sel_FP["SIM_REDSHIFT_CMB"],
            histtype="step",
            color=cm(t),
            linestyle="--",
            bins=bins_to_use,
        )
    ax20.set_xlim(0, 1.2)
    ax20.tick_params(labelsize=14)
    ax20.set_xlabel("simulated redshift", fontsize=18)

    # hist stretch
    n, bins_to_use, tmp = ax21.hist(Ias["x1"], color="black", histtype="step", lw=3)
    for t in threshold_list:
        sel_TP = sel_TP_dic[t]
        ax21.hist(
            sel_TP["x1"],
            orientation="vertical",
            histtype="step",
            color=cm(t),
            bins=bins_to_use,
            lw=2,
        )
    ax21.set_xlabel("x1", fontsize=18)
    ax21.yaxis.set_label_position("right")
    ax21.set_xlim(-3, 3)
    ax21.tick_params(labelsize=14)
    # color histo
    n, bins_to_use, tmp = ax31.hist(Ias["c"], color="black", histtype="step", lw=3)
    for t in threshold_list:
        sel_TP = sel_TP_dic[t]
        ax31.hist(
            sel_TP["c"],
            orientation="vertical",
            histtype="step",
            color=cm(t),
            bins=bins_to_use,
            lw=2,
        )
    ax31.set_xlabel("c", fontsize=18)
    ax31.set_xlim(-1, 1)
    ax31.tick_params(labelsize=14)
    ax31.yaxis.set_label_position("right")

    gs.tight_layout(fig)
    plt.savefig(nameout)
    plt.close()
    del fig


#################
# Formatting
#################


def seaborn_formatting_mag(df, settings):
    """Seaborn friendly formatting

    Basic formatting and eliminating outliers (to avoid rejection by seaborn of pd.DataFrame)

    Args:
        df (DataFrame): predictions
        settings (ExperimentSettings): custom class to hold hyperparameters
    Returns:
        df (DataFrame): reformatted
    """
    df["salt"] = df["dataset_saltfit_2classes"] != -1
    df = du.tag_type(df, settings, type_column=settings.sntype_var)
    # because it doesn't like my normal df
    df_skimmed = pd.DataFrame()
    for f in ["g", "r", "i", "z"]:
        var = "SIM_PEAKMAG_" + f
        df_skimmed[var] = np.array([k for k in df[var].values])
    df_skimmed["salt"] = np.array([k for k in df["salt"].values])
    df_skimmed["target"] = np.array([k for k in df["target_2classes"].values])
    df_skimmed["SIM_REDSHIFT_CMB"] = np.array(
        [k for k in df["SIM_REDSHIFT_CMB"].values]
    )
    df_skimmed[settings.sntype_var] = np.array(
        [k for k in df[settings.sntype_var].values]
    )
    # skimm
    for f in ["g", "r", "i", "z"]:
        var = "SIM_PEAKMAG_" + f
        df_skimmed = df_skimmed[(df_skimmed[var] > 20) & (df_skimmed[var] < 28)]

    return df_skimmed


#################
# Computations
#################


def make_measurements_df(df, settings, group_bayesian=False):
    """Obtain measurements
    Args:
        df (DataFrame): predictions
        settings (ExperimentSettings): custom class to hold hyperparameters
        group_bayesian (Boolean): if BRNNs ar eused need to group predictions
    Returns:
        df (DataFrame): with measurements and necessary keys to compute accuracy
    """
    list_df = []
    for key in ["-2", "-1", "", "+2"]:
        if key != "all":
            format_key = f"PEAKMJD{key}"
        else:
            format_key = key
        tmp_df = pd.DataFrame()
        tmp_df[
            [f"num_{band}" for band in settings.list_filters]
            + [f"{format_key}_class{i}" for i in [0, 1]]
            + ["target", "SNID"]
        ] = df[
            [f"{format_key}_num_{band}" for band in settings.list_filters]
            + [f"{format_key}_class{i}" for i in [0, 1]]
            + ["target", "SNID"]
        ]
        tmp_df = pu.reformat_df(
            tmp_df,
            key,
            settings,
            keep=[f"num_{band}" for band in settings.list_filters],
            group_bayesian=group_bayesian,
        )
        list_df.append(tmp_df)
    ndf = pd.concat(list_df)
    return ndf


def distance_modulus(df):
    """Add distance modulus
    Args:
        df (DataFrame): with SALT2 fitted features
    Returns:
        df (DataFrame): with distance modulus computed
    """

    # SNIa parameters
    Mb = 19.365
    alpha = 0.144  # from sim
    beta = 3.1
    # Add distance modulus to this Data Frame
    df["mu"] = (
        np.array(df["mB"]) + Mb + np.array(alpha * df["x1"]) - np.array(beta * df["c"])
    )
    df["delmu"] = df["mu"].values - dist_mu(df["SIM_REDSHIFT_CMB"].values.astype(float))
    # assuming theoretical mu nor alpha, beta, abs mag have errors
    df["delmu_err"] = (
        np.array(df["mBERR"])
        + np.array(alpha * df["x1ERR"])
        - np.array(beta * df["cERR"])
    )
    return df


def sel_eff(merged, threshold, settings):
    """Efficiency curve for different probabilities

    Args:
        df (DataFrame): with SALT2 fitted features
    Returns:
        mean_bins (np.array)
        efficience (np.array)
    """
    sel = merged[merged["class0"] > threshold].copy()
    if len(sel) > 0:
        mean_bins, efficiency = pu.get_quantity_vs_variable(
            "efficiency", "SIM_REDSHIFT_CMB", sel, settings, nbins=10, mean_bins=True
        )
    else:
        mean_bins, efficiency = np.arange(0, 1.2, 1.2 / 10.0), np.zeros(10)
    return mean_bins, efficiency


#################
# Plot functions
#################


def datasets_plots(df, settings):
    """Dataset violin plots
    peak magnitudes and redshift distributions of representative and non-representative datasets

    Args:
        df (DataFrame): predictions
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    # Reformat into seaborn friendly format
    df = seaborn_formatting_mag(df, settings)
    multiplot_violin_paper(df, "test", settings)


def performance_plots(settings):
    """Performance: accuracy vs. size training set
        Uses only Saltfit data and Baseline and Random Forest models

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    # Read performance summaries (created by supernnova.validate.performance)
    list_files = Path(f"{settings.models_dir}").glob("**/*METRICS*.pickle")
    list_files = map(str, list_files)

    list_files = [
        f
        for f in list_files
        if ("randomforest" in f or "vanilla" in f)
        and "saltfit" in f
        and "global" in f
        and "_CLF_2" in f
        and "0.25" not in f
    ]
    # select only best hp (issue for df 0.2) and only cyclic vanilla
    tmp_list_files = [
        item for item in list_files if "N_global_lstm_32x2_0.05_128_True_mean_C" in item
    ]
    tmp_list_files += [item for item in list_files if "randomforest" in item]
    list_files = tmp_list_files

    if len(list_files) == 0:
        print(
            lu.str_to_brightstr(
                "supernnova/visualization/superNNova_plots:performance_plots.py"
            ),
            lu.str_to_yellowstr(
                "\nNo predictions found. Please train and validate randomforest and vanilla models"
            ),
        )
        return

    # read all performance metrics
    list_df = []
    for f in list_files:
        df = pd.read_pickle(f)
        model_name = df["model_name"][0]
        model_name_noseed = re.sub(r"S\_\d+_", "", model_name)
        df["model_name_noseed"] = model_name_noseed
        # Use regular expressions to look for hyperparameters
        nb_classes = int(re.search(r"(?<=CLF\_)\d+(?=\_)", model_name).group())
        data_fraction = float(re.search(r"(?<=DF\_)\d\.\d+(?=\_)", model_name).group())
        seed = int(re.search(r"(?<=S\_)\d+(?=\_)", model_name).group())
        redshift = re.search(r"(?<=R\_)[A-Za-z]+(?=\_)", model_name).group()
        model_type = model_name.split("_")[0]
        model_source_data = re.search(r"(?<=\_)[A-Za-z]+(?=\_DF\_)", model_name).group()
        # avoiding models trained and validated with different datasets
        if model_source_data != df["source_data"].values.astype(str)[0]:
            continue

        df["model_type"] = model_type
        df["data_fraction"] = data_fraction
        df["nb_classes"] = nb_classes
        df["seed"] = seed
        df["redshift"] = redshift

        list_df.append(df)

    df_all = pd.concat(list_df, axis=0, sort=True)

    # Plot acc vs. n SN in training
    plot_acc_vs_nsn(df_all, settings)


def purity_vs_z(df, model_name, settings):
    """Purity and contamination redshift evolution

    Args:
        df (DataFrame):
        modelname (str): name of model to be used
        settings (ExperimentSettings): custom class to hold hyperparameters
    """
    df = pu.reformat_df(df, "all", settings, keep=None, group_bayesian=True)

    bin_centers, purity_arr = pu.get_quantity_vs_variable(
        "purity", "SIM_REDSHIFT_CMB", df, settings
    )
    y_dic = {}
    y_dic["purity"] = purity_arr
    for typ in [t for t in settings.sntypes if t != 101]:
        bin_centers, y_dic[typ] = pu.get_quantity_vs_variable(
            "contamination", "SIM_REDSHIFT_CMB", df, settings, contamination_by=typ
        )

    # now make the plots for each model
    plot_path = f"{settings.figures_dir}/puritycontamination_vs_z"
    os.makedirs(plot_path, exist_ok=True)
    nameout = f"{plot_path}/{model_name}.png"
    color_sequence = [ALL_COLORS[4]] + [
        ALL_COLORS[i] for i in range(len(y_dic)) if i != 4
    ]
    label_list = ["purity"] + [
        f"{settings.sntypes[k]}" for k in y_dic.keys() if k != 101 and k != "purity"
    ]
    binned_2d(
        bin_centers,
        y_dic,
        "redshift",
        "purity/contamination",
        label_list,
        color_sequence,
        MARKER_LIST,
        nameout,
        extra_line=True,
    )


def cadence_acc_matrix(df, model_name, settings):
    """Matrix with accuracy w.r. number of measurements in a band

    Correlation between accuracy and a certain number of observations required per filter

    Args:
        df (DataFrame):
        modelname (str): name of model to be used
        settings (ExperimentSettings): custom class to hold hyperparameters
    """
    if "bayesian" or "variational" in model_name:
        measurements_df = make_measurements_df(df, settings, group_bayesian=True)
    else:
        measurements_df = make_measurements_df(df, settings)
    n_measurements_dic = {}
    accuracy_dic = {}
    for band in settings.list_filters:
        n_measurements_dic[band], accuracy_dic[band] = pu.get_quantity_vs_variable(
            "accuracy",
            f"num_{band}",
            measurements_df,
            settings,
            nbins=max(measurements_df[f"num_{band}"]),
            intervals=True,
        )
    plot_path = f"{settings.figures_dir}/accuracy_measurements_matrix"
    os.makedirs(plot_path, exist_ok=True)
    nameout = f"{plot_path}/{model_name}.png"
    plot_acc_matrix(accuracy_dic, n_measurements_dic, settings, nameout)


def hubble_residuals(df, model_name, fits, settings):
    """Hubble residuals for classified supernovae as type Ia

    Uses SALT2 fits to compute the distance modulus, therefore list is not complete if fit failed.

    Args:
        df (DataFrame):
        modelname (str): name of model to be used
        fits (DataFrame): information from SALT2 fit
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    # Init plot
    plot_path = f"{settings.figures_dir}/HDresiduals"
    os.makedirs(plot_path, exist_ok=True)

    def reformat_df_HR(df, model_name, settings, keep_list=["mB", "x1", "c"]):
        """format and add distance modulus info
        including error bars
        """
        df = pu.reformat_df(df, "all", settings, keep=keep_list, group_bayesian=True)
        df = pd.merge(df, fits, on="SNID")
        df = distance_modulus(df)
        df = df[~np.isnan(df["mu"])]
        return df

    # Hresiduals with probability thresholds
    merged = reformat_df_HR(df, model_name, settings, keep_list=["mB", "x1", "c"])
    nameout = f"{plot_path}/{model_name}.png"
    plot_HDres_histos_vs_z(
        merged,
        nameout,
        threshold_var="class0",
        threshold_list=[0.5, 0.7, 0.9],
        threshold_sign=">",
        settings=settings,
    )

    # residuals with uncertainty thresholds
    if "bayesian" in model_name or "variational" in model_name:
        # get one std deviation
        df_grouped = df.groupby("SNID", as_index=False)
        sigma_df = create_sigma_df(df_grouped)
        df_tmp = pd.merge(df, sigma_df, on="SNID")
        merged = reformat_df_HR(
            df_tmp, model_name, settings, keep_list=["mB", "x1", "c", "sigma_all"]
        )
        nameout = f"{plot_path}/{model_name}_cut_uncertainty.png"
        plot_HDres_histos_vs_z(
            merged,
            nameout,
            threshold_var="sigma_all",
            threshold_list=[0.05, 0.1, 0.2],
            threshold_sign="<",
            settings=settings,
        )


def cnf_matrix(df, model_name, settings):
    """Get confusion matrix from predictions

    Args:
        df (DataFrame): predictions for a given model
        modelname (str): name of model to be used
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    df_r = pu.reformat_df(df, "all", settings)

    # Get sorted unique target values (sklearn confusion_matrix sorts labels)
    unique_targets = sorted(df_r["target"].unique())
    cnf_matrix = confusion_matrix(
        df_r["target"], df_r["predicted_target"], labels=unique_targets
    )

    plt.figure()
    settings.nb_classes = len(unique_targets)
    class_names = [
        du.sntype_decoded(i, settings, simplify=True) for i in unique_targets
    ]
    plot_confusion_matrix(
        settings, cnf_matrix, classes=class_names, normalize=True, nameout=model_name
    )


def plot_speed_benchmark(dump_dir):
    """Plot RNN inference speed benchmarks

    **N.B.** You should have run the speed benchmarks with

    .. code: python

        python run.py --speed

    - Plot speed based on device
    - Plot speed based on choice of model (BBB, Variational or Baseline)


    Args:
        dump_dir (str): Root folder where results are stored
    """

    speed_file = os.path.join(dump_dir, "stats/rnn_speed.csv")

    assert os.path.isfile(speed_file), lu.str_to_redstr(
        "speed_file does not exist. Run ``python run.py --speed`` first."
    )

    df = pd.read_csv(speed_file)

    df_cpu = df[df.device == "cpu"]
    df_gpu = df[df.device == "gpu"]

    cpu_is_available = len(df_cpu) > 0
    gpu_is_available = len(df_gpu) > 0

    # CPU benchmark should always be available
    assert cpu_is_available

    n_models = len(df.model.unique())

    if gpu_is_available:
        # Space bars by 2 units to leave room for gpu
        idxs_cpu = 0.5 + np.arange(3 * n_models)[::3]
        idxs_gpu = idxs_cpu + 1
        xticks = idxs_cpu + 0.5
        xtick_labels = df_cpu.model.values.tolist()

    else:
        # Space bars by 1 unit
        idxs_cpu = 0.5 + np.arange(2 * n_models)[::2]
        xticks = idxs_cpu
        xtick_labels = df_cpu.model.values.tolist()

    plt.figure()
    ax = plt.gca()

    for i in range(len(idxs_cpu)):
        label = "CPU" if i == 0 else None
        plt.bar(
            idxs_cpu[i],
            df_cpu["Supernova_per_s"].values[i],
            width=1,
            color="C0",
            label=label,
        )

    if gpu_is_available:
        for i in range(len(idxs_gpu)):
            label = "GPU" if i == 0 else None
            plt.bar(
                idxs_gpu[i],
                df_gpu["Supernova_per_s"].values[i],
                width=1,
                color="C2",
                label=label,
            )

    ax.set_ylabel("Lightcurves / s", fontsize=16)
    ax.set_title("Inference throughput", fontsize=20)

    ax.set_xticks(xticks)
    ax.set_xticklabels(xtick_labels)
    ax.set_yscale("log")
    ax.legend()

    plt.grid()
    plt.tight_layout()
    plt.savefig(os.path.join(dump_dir, "figures/rnn_speed.png"))
    plt.clf()
    plt.close()


def create_OOD_classification_plots(df, list_models, settings):
    """OOD classification for a model (ppercentages of predictions by class and uncertainties)

    Args:
        df (pandas.DataFrame) : summary statistics df
        list: list with models to use
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    def OOD_classification_percentages(df, model, targets=2):

        df_sel = df.copy()
        df_sel = df_sel[
            df_sel["model_name_noseed"] == model.replace("CLF_2", f"CLF_{targets}")
        ]
        df_sel = df_sel.fillna(0)

        if len(df_sel) != 0:
            # Get the number of objects classified into a certain class
            # compute percentages
            list_keys = [
                k
                for k in df_sel.keys()
                if "all_" in k and "_num_pred_class" in k and "mean" in k
            ]
            # hack to get total number of OOD lcs (they should be all the same)
            df_perc = (
                df_sel[list_keys]
                .div(
                    df_sel[[item for item in list_keys if "random" in item]].sum(
                        axis=1
                    ),
                    axis=0,
                )
                .multiply(100)
                .round(2)
            )

            # add percentages of non-classifications
            df_perc["all_percentage_non_pred_mean"] = df_sel[
                "all_percentage_non_pred_mean"
            ]
            for prefix in ["random", "shuffle", "reverse", "sin"]:
                df_perc[f"all_{prefix}_percentage_non_pred_mean"] = df_sel[
                    f"all_{prefix}_percentage_non_pred_mean"
                ]

            # add model name
            df_perc["modelname"] = df_sel["model_name_noseed"].apply(
                lambda x: x.split("_")[1]
            )
            df_perc = df_perc.replace("vanilla", "baseline")
            df_perc = df_perc.fillna(0).round(2)

            # Group by target type: SNe or OOD types
            # save in dictionary for plotting
            percentages = {}
            for to_classify in ["SNe", "random", "shuffle", "reverse", "sin"]:
                if to_classify != "SNe":
                    key_list = [
                        f"all_{to_classify}_num_pred_class{i}_mean"
                        for i in range(targets)
                    ] + [f"all_{to_classify}_percentage_non_pred_mean"]
                else:
                    key_list = [
                        f"all_num_pred_class{i}_mean" for i in range(targets)
                    ] + ["all_percentage_non_pred_mean"]
                percentages[to_classify] = df_perc[key_list].values.flatten().tolist()

        else:
            percentages = {}
            for to_classify in ["SNe", "random", "shuffle", "reverse", "sin"]:
                percentages[to_classify] = np.zeros(targets + 1).tolist()
        return percentages

    def autolabel(ax, rects):
        """
        Attach a text label above each bar displaying its height
        """
        for rect in rects:
            height = rect.get_height()
            if height > 90:
                factor_text = 0.8
            else:
                factor_text = 1.05
            ax.text(
                rect.get_x() + rect.get_width() / 2.0,
                (factor_text * height),
                f"{height}",
                ha="center",
                va="bottom",
                fontsize=32,
            )

    def barplot_single(ax, labels, percentage, color):
        rects = ax.bar(
            np.arange(len(labels)),
            percentage,
            color=color,
            edgecolor="black",
            tick_label=labels,
        )
        ax.yaxis.set_tick_params(labelsize=40)
        ax.xaxis.set_tick_params(labelsize=40)
        ax.set_ylim(0.01, 99.9)
        autolabel(ax, rects)

    for model in list_models:
        # fig, ax = plt.subplots(5, 3, figsize=(25, 30), sharey=True)
        fig = plt.figure(figsize=(25, 30))
        gs = gridspec.GridSpec(
            5, 3, height_ratios=[1, 1, 1, 1, 1], width_ratios=[1, 1, 2]
        )
        ax = {}
        ax["00"] = plt.subplot(gs[0])
        ax["01"] = plt.subplot(gs[1], sharey=ax["00"])
        ax["02"] = plt.subplot(gs[2], sharey=ax["00"])
        ax["10"] = plt.subplot(gs[3])
        ax["11"] = plt.subplot(gs[4], sharey=ax["10"])
        ax["12"] = plt.subplot(gs[5], sharey=ax["10"])
        ax["20"] = plt.subplot(gs[6])
        ax["21"] = plt.subplot(gs[7], sharey=ax["20"])
        ax["22"] = plt.subplot(gs[8], sharey=ax["20"])
        ax["30"] = plt.subplot(gs[9])
        ax["31"] = plt.subplot(gs[10], sharey=ax["30"])
        ax["32"] = plt.subplot(gs[11], sharey=ax["30"])
        ax["40"] = plt.subplot(gs[12])
        ax["41"] = plt.subplot(gs[13], sharey=ax["40"])
        ax["42"] = plt.subplot(gs[14], sharey=ax["40"])

        # Percentages for a predicted class
        for j, target in enumerate([2, 3, len(settings.sntypes)]):
            percentages = OOD_classification_percentages(df, model, targets=target)
            labels = []
            for t in range(target):
                settings.nb_classes = target
                typ = du.sntype_decoded(t, settings)
                labels.append(typ.strip("SN "))
            # add non-classified
            labels.append("None")
            colors = ALL_COLORS[:target] + ["gray"]
            # Row
            for i, to_classify in enumerate(
                ["SNe", "random", "shuffle", "reverse", "sin"]
            ):
                barplot_single(ax[f"{i}{j}"], labels, percentages[to_classify], colors)
                if i == 0:
                    ax[f"{i}{j}"].set_title(class_target_decode(target), fontsize=50)
                if i < 4:
                    plt.setp(ax[f"{i}{j}"].get_xticklabels(), visible=False)
                ax[f"{i}0"].set_ylabel(to_classify, fontsize=50)
                # dont show axis
                if i < 4:
                    plt.setp(ax[f"{i}{j}"].get_xticklabels(), visible=False)
                if j > 0:
                    plt.setp(ax[f"{i}{j}"].get_yticklabels(), visible=False)
        plt.subplots_adjust(
            left=0.08, right=0.99, bottom=0.03, top=0.95, wspace=0.0, hspace=0.02
        )
        # fig.tight_layout()
        plot_path = f"{settings.figures_dir}/OOD_percentages"
        Path(plot_path).mkdir(parents=True, exist_ok=True)
        fig.savefig(f"{plot_path}/OOD_{model}_classification_percentages.png")
        plt.close()
        plt.clf()


def science_plots(settings, onlycnf=False):
    """Plots for SuperNNova paper

    Saved in settings.figures_dir. Need to provide prediction files and linked settings

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters

    """

    if len(settings.prediction_files) == 0:
        print(
            lu.str_to_yellowstr("Warning: no prediction files provided. Not plotting")
        )
        return

    # Load data
    df_SNinfo = du.load_HDF5_SNinfo(settings)

    # check if files are there
    tmp_not_found = [m for m in settings.prediction_files if not os.path.exists(m)]
    if len(tmp_not_found) > 0:
        print(lu.str_to_redstr(f"Files not found {tmp_not_found}"))
        tmp_prediction_files = [
            m for m in settings.prediction_files if os.path.exists(m)
        ]
        settings.prediction_files = tmp_prediction_files

    for f in settings.prediction_files:
        if Path(f).suffix == ".pickle":
            df = pd.read_pickle(f)
        elif Path(f).suffix == ".csv":
            df = pd.read_csv(f)
            df["SNID"] = df["SNID"].astype(str)
        model_name = Path(f).stem

        cols_to_merge = [
            "SNID",
            "SIM_REDSHIFT_CMB",
            settings.sntype_var,
            "mB",
            "x1",
            "c",
        ]
        cols_to_merge += [c for c in df_SNinfo.columns if "unique_nights" in c]
        cols_to_merge += [c for c in df_SNinfo.columns if "_num_" in c]

        df = df.merge(df_SNinfo.reset_index()[cols_to_merge], how="left", on="SNID")

        if onlycnf:
            cnf_matrix(df, model_name, settings)
        else:
            # Science plots
            purity_vs_z(df, model_name, settings)
            # cadence_acc_matrix(df, model_name, settings)
            # Get extra info from fits (for distance modulus)
            fits = du.load_fitfile(settings)
            if len(fits) != 0:
                fits = fits[["SNID", "cERR", "mBERR", "x1ERR"]]
                hubble_residuals(df, model_name, fits, settings)
            cnf_matrix(df, model_name, settings)



import os
import h5py
import numpy as np
import pandas as pd
from tqdm import tqdm
from astropy.table import Table
from collections import namedtuple

from . import logging_utils

OFFSETS = [-2, -1, 0, 1, 2]
OOD_TYPES = ["random", "reverse", "shuffle", "sin"]
OFFSETS_STR = ["-7", "-2", "-1", "", "+1", "+2", "+30"]
OFFSETS_VAL = []
for v in OFFSETS_STR:
    if v == "":
        OFFSETS_VAL.append(0)
    else:
        OFFSETS_VAL.append(int(v.replace("+", "")))

LogStandardized = namedtuple("LogStandardized", ["arr_min", "arr_mean", "arr_std"])


def load_pandas_from_fit(fit_file_path):
    """Load a FIT file and cast it to a PANDAS dataframe

    Args:
        fit_file_path (str): path to FIT file

    Returns:
        (pandas.DataFrame) load dataframe from FIT file
    """
    dat = Table.read(fit_file_path, format="fits")
    df = dat.to_pandas()

    return df


def sntype_decoded(target, settings, simplify=False):
    """Match the target class (integer in {0, ..., 6} to the name
    of the class, i.e. something like "SN Ia" or "SN CC"

    Args:
        target (int): specifies the classification target
        settings (ExperimentSettings): custom class to hold hyperparameters
        simplify (Boolean): if True do not show all classes

    Returns:
        (str) the name of the class

    """
    target_class = getattr(settings, "target_sntype", "Ia")

    # Build ordered unique class list consistent with tag_type:
    # target_sntype first, then remaining classes in insertion order
    used = set()
    unique_classes = [
        x
        for x in settings.sntypes.values()
        if x not in used and (used.add(x) or True)
    ]
    if target_class in unique_classes:
        unique_classes.remove(target_class)
        unique_classes.insert(0, target_class)

    if settings.nb_classes > 2:
        if target < len(unique_classes):
            SNtype = unique_classes[target]
        else:
            SNtype = f"class_{target}"
    else:
        if target == 0:
            if target_class in unique_classes:
                SNtype = f"SN {target_class}"
            else:
                SNtype = f"SN {unique_classes[0]}"
        else:
            non_target = [k for k in unique_classes if k != target_class]
            SNtype = f"SN {'|'.join(non_target)}"
            if simplify:
                if target_class in unique_classes:
                    SNtype = f"non SN {target_class}"
                else:
                    SNtype = f"non SN {unique_classes[0]}"
    return SNtype


def tag_type(df, settings, type_column="TYPE"):
    """Create classes based on a type columns

    Depending on the number of classes (2 or all), we create distinct
    target columns

    Args:
        df (pandas.DataFrame): the input dataframe
        settings (ExperimentSettings): controls experiment hyperparameters
        type_column (str): the type column in df

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # SNTYPE checks
    if type_column not in df.keys():
        if settings.data_testing:
            df[settings.sntype_var] = np.ones(len(df)).astype(int)
        else:
            logging_utils.print_red(
                "Please provide SNTYPE with data (else use data_testing option)"
            )
            raise Exception

    # Auto-detect types in data not present in sntypes and assign as contaminant
    df[type_column] = df[type_column].astype(str)
    all_types_in_data = set(df[type_column].unique())
    missing_types = [t for t in all_types_in_data if t not in settings.sntypes]
    if len(missing_types) > 0:
        logging_utils.print_yellow(
            "Missing sntypes",
            f"{missing_types} assigned to 'contaminant' class",
        )
        for mtyp in missing_types:
            settings.sntypes[mtyp] = "contaminant"

    # 2 classes: target_sntype vs rest
    target_class = getattr(settings, "target_sntype", "Ia")
    list_types = list(set([x for x in settings.sntypes.values()]))
    if target_class in list_types:
        keys_target = [
            key for (key, value) in settings.sntypes.items() if value == target_class
        ]
        df["target_2classes"] = df[type_column].apply(
            lambda x: 0 if x in keys_target else 1
        )
    else:
        # Fallback: first key in sntypes = class 0
        logging_utils.print_yellow(
            "target_sntype",
            f"'{target_class}' not found in sntypes values {list_types}, "
            f"using first entry as class 0",
        )
        first_key = list(settings.sntypes.keys())[0]
        df["target_2classes"] = df[type_column].apply(
            lambda x: 0 if x == first_key else 1
        )

    # All classes — ensure target_sntype is class 0 for consistency with binary
    used = set()
    unique_classes = [
        x for x in settings.sntypes.values() if x not in used and (used.add(x) or True)
    ]
    if target_class in unique_classes:
        unique_classes.remove(target_class)
        unique_classes.insert(0, target_class)

    classes_to_use = dict([(y, x) for x, y in enumerate(unique_classes)])
    map_keys_to_classes = {}
    for k, v in settings.sntypes.items():
        map_keys_to_classes[k] = classes_to_use[v]

    df[f"target_{len(unique_classes)}classes"] = df[type_column].apply(
        lambda x: map_keys_to_classes[x]
    )

    return df


def load_fitfile(settings, verbose=True):
    """Load the FITOPT file as a pandas dataframe

    Pickle it for future use (it is faster to load as a pickled dataframe)

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        verbose (bool): whether to display logging message. Default: ``True``

    Returns:
        (pandas.DataFrame) dataframe with FITOPT data
    """
    if verbose:
        logging_utils.print_green("Loading FITRES file")

    if os.access(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle", os.R_OK):
        df = pd.read_pickle(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle")
        if verbose:
            print(f"Loaded {settings.preprocessed_dir}/FITOPT000.FITRES.pickle")

    elif os.access(f"{settings.fits_dir}/FITOPT000.FITRES", os.R_OK) or os.access(
        f"{settings.fits_dir}/FITOPT000.FITRES.gz", os.R_OK
    ):
        fit_name = (
            f"{settings.fits_dir}/FITOPT000.FITRES"
            if os.access(f"{settings.fits_dir}/FITOPT000.FITRES", os.R_OK)
            else f"{settings.fits_dir}/FITOPT000.FITRES.gz"
        )
        df = pd.read_csv(
            fit_name, index_col=False, comment="#", delimiter=" ", skipinitialspace=True
        )
        df = tag_type(df, settings)

        # Rename CID to SNID
        # SNID is CID in FITOPT000.FITRES
        if "SNID" not in df.keys():
            df = df.rename(columns={"CID": "SNID"})

        # Save to pickle for later use and fast reload
        df.to_pickle(f"{settings.preprocessed_dir}/FITOPT000.FITRES.pickle", protocol=4)
        if verbose:
            print(f"Loaded {fit_name}")
    else:
        # returning empty df
        df = pd.DataFrame()
        if verbose:
            logging_utils.print_yellow("Warning: No FITRES file to load")

    return df


def process_header_FITS(file_path, settings, columns=None):
    """Read the HEAD FIT file, add target columns and return
    in pandas DataFrame format

    Args:
        file_path (str): the  path to the header FIT file
        settings (ExperimentSettings): controls experiment hyperparameters
        columns (lsit): list of columns to keep. Default: ``None``

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # Data
    df = load_pandas_from_fit(file_path)

    try:
        df["SNID"] = df["SNID"].str.decode("utf-8")
    except Exception:
        df["SNID"] = df["SNID"].astype(str)

    df[settings.sntype_var] = df[settings.sntype_var].astype(str)

    df = tag_type(df, settings, type_column=settings.sntype_var)

    if columns is not None:
        df = df[columns]

    return df


def process_header_csv(file_path, settings, columns=None):
    """Read the HEAD csv file, add target columns and return
    in pandas DataFrame format

    Args:
        file_path (str): the  path to the header FIT file
        settings (ExperimentSettings): controls experiment hyperparameters
        columns (lsit): list of columns to keep. Default: ``None``

    Returns:
        (pandas.DataFrame) the dataframe, with new target columns
    """

    # Data
    df = pd.read_csv(file_path)
    df = tag_type(df, settings, type_column=settings.sntype_var)

    if columns is not None:
        df = df[columns]
    return df


# TODO: double-check the following function that been commented out before deleting. It seems only related to randomforest.
# def add_redshift_features(settings, df):
#     """Add redshift features to pandas dataframe.

#     Args:
#         settings (ExperimentSettings): controls experiment hyperparameters
#         df (str): pandas DataFrame with FIT data

#     Returns:
#         (pandas.DataFrame) the dataframe, possibly with added redshift features
#     """

#     # check if we use host redshift as feature
#     host_features = [f for f in settings.randomforest_features if "HOST" in f]
#     use_redshift = len(host_features) > 0

#     if use_redshift > 0:
#         logging_utils.print_green("Adding redshift features...")

#         columns_to_read = ["SNID"] + host_features
#         # reading from batch pickles
#         list_files = natsorted(glob.glob(f"{settings.preprocessed_dir}/*_PHOT.pickle"))
#         # Check file with redshift features exist
#         error_msg = "Preprocessed_file not found. Call python run.py --data"
#         assert os.path.isfile(list_files[0]), error_msg

#         extra_info_df = pd.concat(
#             [pd.read_pickle(f)[columns_to_read] for f in list_files]
#         )

#         # In extra_info_df, there are many SNID duplicates as each row corresponds to a time step in a given curve
#         # We use groupby + first to only select the first row of each lightcurve
#         # Then we can merge knowing there won't be SNID duplicates in extra_info_df
#         extra_info_df = (
#             extra_info_df.groupby("SNID")[host_features].first().reset_index()
#         )

#         # Add redshift info to df
#         df = df.merge(extra_info_df, how="left", on="SNID")

#     return df


def compute_delta_time(df):
    """Compute the delta time between two consecutive observations

    Args:
        df (pandas.DataFrame): dataframe holding lightcurve data

    Returns:
        (pandas.DataFrame) dataframe holding lightcurve data with delta_time features
    """
    # in case the photometyr is not time sorted
    df = df.sort_values(["SNID", "MJD"])
    df["delta_time"] = df["MJD"].diff()
    # Fill the first row with 0 to replace NaN
    df.delta_time = df.delta_time.fillna(0)
    try:
        IDs = df.SNID.values
    # Deal with the case where lightcrv_ID is the index
    except AttributeError:
        assert df.index.name == "SNID"
        IDs = df.index.values
    # Find idxs of rows where a new light curve start then zero delta_time
    idxs = np.where(IDs[:-1] != IDs[1:])[0] + 1
    arr_delta_time = df.delta_time.values
    arr_delta_time[idxs] = 0
    df["delta_time"] = arr_delta_time

    return df


def remove_data_post_large_delta_time(df):
    """
    Remove rows in the same light curve after a gap > 150 days
    Reason: If no signal has been saved in a time frame of 150 days,
    it is unlikely there is much left afterwards

    Args:
        df (pandas.DataFrame): dataframe holding lightcurve data

    Returns:
        (pandas.DataFrame) dataframe where large delta time rows have been removed
    """

    # Identify indices where delta time is large
    list_to_remove = []
    idx_high = np.where(df.delta_time.values > 150)[0]
    # Identify the lightcurve ID where this happens
    IDs = df.SNID.values
    # Loop over indices and remove row if they belong to the same light curve
    for idx in idx_high:
        ID = IDs[idx]
        same_lc = True
        while same_lc:
            list_to_remove.append(idx)
            idx += 1
            new_ID = IDs[idx]
            same_lc = new_ID == ID
    df = df.drop(list_to_remove)
    # Reset index to account for dropped rows
    df = df.reset_index(drop=True)

    return df


def load_HDF5_SNinfo(settings):
    """Load physical information related to the created database of lightcurves

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters

    Returns:
        (pandas.DataFrame) dataframe holding physics information about the dataset
    """

    file_name = f"{settings.processed_dir}/database.h5"

    dict_SNinfo = {}
    with h5py.File(file_name, "r") as hf:

        columns_to_keep = ["SNID", settings.sntype_var, "mB", "c", "x1"]

        columns_to_keep += [c for c in hf.keys() if "SIM_" in c]
        columns_to_keep += [c for c in hf.keys() if "dataset_" in c]
        columns_to_keep += [c for c in hf.keys() if "PEAK" in c]

        for key in columns_to_keep:
            dict_SNinfo[key] = hf[key][:]

    df_SNinfo = pd.DataFrame(dict_SNinfo)
    # bytes
    if isinstance(df_SNinfo["SNID"].values[0], bytes):
        df_SNinfo["SNID"] = df_SNinfo["SNID"].str.decode("utf8")
    return df_SNinfo


def log_standardization(arr):
    """Normalization strategy for the fluxes and fluxes error

    - Log transform the data
    - Mean and std dev normalization

    Args:
        arr (np.array): data to normalize

    Returns:
        (LogStandardized) namedtuple holding normalization data
    """

    arr_min = np.min(arr)
    arr_log = np.log(-arr_min + arr + 1e-5)
    arr_mean = arr_log.mean()
    arr_std = arr_log.std()

    if arr_min < -2000:
        logging_utils.print_yellow(
            f"Warning: extreme data values {arr_min}",
            "clipping normalization min to -2000",
        )
        arr_min = -2000

    return LogStandardized(arr_min=arr_min, arr_mean=arr_mean, arr_std=arr_std)


def save_to_HDF5(settings, df):
    """Saved processed dataframe to HDF5

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
        df (pandas.DataFrame): dataframe holding processed data

    """
    # One hot encode filter information and Normalize features
    list_training_features = [f"FLUXCAL_{f}" for f in settings.list_filters]
    list_training_features += [f"FLUXCALERR_{f}" for f in settings.list_filters]
    list_training_features += [
        "delta_time",
        "HOSTGAL_PHOTOZ",
        "HOSTGAL_PHOTOZ_ERR",
        "HOSTGAL_SPECZ",
        "HOSTGAL_SPECZ_ERR",
    ]
    if settings.additional_train_var:
        list_training_features += list(settings.additional_train_var)

    list_misc_features = [
        "PEAKMJD",
        settings.sntype_var,
        "mB",
        "c",
        "x1",
        "SIM_REDSHIFT_CMB",
        "SIM_PEAKMAG_z",
        "SIM_PEAKMAG_g",
        "SIM_PEAKMAG_r",
        "SIM_PEAKMAG_i",
    ]

    if settings.photo_window_var not in list_misc_features:
        list_misc_features += settings.photo_window_var

    list_misc_features = [k for k in list_misc_features if k in df.keys()]

    assert df.index.name == "SNID", "Must set SNID as index"

    # Get the list of lightcurve IDs
    ID = df.index.values
    # Find out when ID changes => find start and end idx of each lightcurve
    idx_change = np.where(ID[1:] != ID[:-1])[0] + 1
    idx_change = np.hstack(([0], idx_change, [len(df)]))
    list_start_end = [(s, e) for s, e in zip(idx_change[:-1], idx_change[1:])]
    # N.B. We could use df.loc[SNID], more elegant but much slower

    # Filter list start end so we get only light curves with at least 3 points
    # except when creating testing data (we want to classify all lcs even w. 1-2 epochs)
    if not settings.data_testing:
        list_start_end = list(filter(lambda x: x[1] - x[0] >= 3, list_start_end))

    # Shuffle
    np.random.shuffle(list_start_end)

    # Save hdf5 file
    with h5py.File(settings.hdf5_file_name, "w") as hf:

        n_samples = len(list_start_end)
        used = set()
        unique_classes = [
            x
            for x in settings.sntypes.values()
            if x not in used and (used.add(x) or True)
        ]
        list_classes = list(set([2, len(unique_classes)]))
        list_names = ["target", "dataset_photometry", "dataset_saltfit"]

        # These arrays can be filled in one shot
        start_idxs = [i[0] for i in list_start_end]
        shuffled_ID = ID[start_idxs]
        hf.create_dataset("SNID", data=shuffled_ID, dtype=h5py.special_dtype(vlen=str))
        df_SNID = pd.DataFrame(shuffled_ID, columns=["SNID"])
        logging_utils.print_green("Saving misc features")
        for feat in list_misc_features:
            if feat == settings.sntype_var:
                dtype = np.dtype("int32")
            else:
                dtype = np.dtype("float32")
            hf.create_dataset(feat, data=df[feat].values[start_idxs], dtype=dtype)
            df = df.drop(columns=feat)

        logging_utils.print_green("Saving class")
        for c_ in list_classes:
            for name in list_names:
                field_name = f"{name}_{c_}classes"
                hf.create_dataset(
                    field_name,
                    data=df[field_name].values[start_idxs],
                    dtype=np.dtype("int8"),
                )
                df = df.drop(columns=field_name)

        df["time"] = df[["delta_time"]].groupby(df.index).cumsum()
        df = df.reset_index()

        logging_utils.print_green("Saving unique nights")
        # Compute how many unique nights of data taking existed around PEAKMJD
        for offset, suffix in zip(OFFSETS, OFFSETS_STR):
            new_column = f"PEAKMJD{suffix}_unique_nights"
            df_nights = (
                df[df["time"] < df["PEAKMJDNORM"] + offset][["PEAKMJDNORM", "SNID"]]
                .groupby("SNID")
                .count()
                .astype(np.uint8)
                .rename(columns={"PEAKMJDNORM": new_column})
                .reset_index()
            )

            hf.create_dataset(
                new_column,
                data=df_SNID.merge(df_nights, on="SNID", how="left")[new_column].values,
                dtype=np.dtype("uint8"),
            )

        logging_utils.print_green("Saving filter occurences")
        # Compute how many occurences of a specific filter around PEAKMJD
        for flt in settings.list_filters:
            # Check presence / absence of the filter at all time steps
            df[f"has_{flt}"] = df.FLT.str.contains(flt).astype(np.uint8)
            for offset, suffix in zip(OFFSETS, OFFSETS_STR):
                new_column = f"PEAKMJD{suffix}_num_{flt}"
                df_flt = (
                    df[df["time"] < df["PEAKMJDNORM"] + offset][[f"has_{flt}", "SNID"]]
                    .groupby("SNID")
                    .sum()
                    .astype(np.uint8)
                    .rename(columns={f"has_{flt}": new_column})
                    .reset_index()
                )
                hf.create_dataset(
                    new_column,
                    data=df_SNID.merge(df_flt, on="SNID", how="left")[
                        new_column
                    ].values,
                    dtype=np.dtype("uint8"),
                )

            df = df.drop(columns=f"has_{flt}")

        # FInally save PEAKMJDNORM
        hf.create_dataset(
            "PEAKMJDNORM",
            data=df["PEAKMJDNORM"].values[start_idxs],
            dtype=np.dtype("float32"),
        )
        cols_to_drop = [
            k
            for k in ["time", "SNID", "PEAKMJDNORM", settings.photo_window_var]
            if k in df.keys()
        ]
        df = df.drop(columns=list(set(cols_to_drop)))

        ########################
        # Normalize per feature
        ########################
        logging_utils.print_green("Compute normalizations")
        gnorm = hf.create_group("normalizations")

        # using normalization per feature
        for feat in settings.training_features_to_normalize:
            # Log transform plus mean subtraction and standard dev subtraction
            log_standardized = log_standardization(df[feat].values)
            # Store normalization parameters
            gnorm.create_dataset(f"{feat}/min", data=log_standardized.arr_min)
            gnorm.create_dataset(f"{feat}/mean", data=log_standardized.arr_mean)
            gnorm.create_dataset(f"{feat}/std", data=log_standardized.arr_std)

        #####################################
        # Normalize flux and fluxerr globally
        #####################################
        logging_utils.print_green("Compute global normalizations")
        gnorm = hf.create_group("normalizations_global")

        ################
        # FLUX features
        #################
        flux_features = [f"FLUXCAL_{f}" for f in settings.list_filters]
        flux_log_standardized = log_standardization(df[flux_features].values)
        # Store normalization parameters
        gnorm.create_dataset("FLUXCAL/min", data=flux_log_standardized.arr_min)
        gnorm.create_dataset("FLUXCAL/mean", data=flux_log_standardized.arr_mean)
        gnorm.create_dataset("FLUXCAL/std", data=flux_log_standardized.arr_std)

        ###################
        # FLUXERR features
        ###################
        fluxerr_features = [f"FLUXCALERR_{f}" for f in settings.list_filters]
        fluxerr_log_standardized = log_standardization(df[fluxerr_features].values)
        # Store normalization parameters
        gnorm.create_dataset("FLUXCALERR/min", data=fluxerr_log_standardized.arr_min)
        gnorm.create_dataset("FLUXCALERR/mean", data=fluxerr_log_standardized.arr_mean)
        gnorm.create_dataset("FLUXCALERR/std", data=fluxerr_log_standardized.arr_std)

        ####################################
        # Save the rest of the data to hdf5
        ####################################

        logging_utils.print_green("Save non-data features to HDF5")

        # This type allows one to store flat arrays of variable
        # length inside an HDF5 group
        data_type = h5py.special_dtype(vlen=np.dtype("float32"))

        hf.create_dataset("data", (n_samples,), dtype=data_type)

        # If header does not have HOST info fill with empty arrays
        list_to_fill = [
            k for k in list_training_features if k not in df.columns.values.tolist()
        ]
        if len([k for k in list_to_fill if "HOST" not in k]) > 0:
            logging_utils.print_red("missing information in input")
            raise AttributeError
        for key in list_to_fill:
            df[key] = np.zeros(len(df))

        logging_utils.print_green("Fit onehot on FLT")
        assert sorted(df.columns.values.tolist()) == sorted(
            list_training_features + ["FLT"]
        )

        # Fit a one hot encoder for FLT
        # to have the same onehot for all datasets
        # tmp = pd.Series(settings.list_filters_combination).append(df["FLT"])
        tmp = pd.concat(
            [pd.Series(settings.list_filters_combination), df["FLT"]]
        )  # TODO: NEED TO TEST THIS LINE
        tmp_onehot = pd.get_dummies(tmp)
        # this is ok since it goes by length not by index (which I never reset)
        FLT_onehot = tmp_onehot[len(settings.list_filters_combination) :]
        df = pd.concat([df[list_training_features], FLT_onehot], axis=1)
        # store feature names
        list_training_features = df.columns.values.tolist()
        hf.create_dataset(
            "features",
            (len(list_training_features),),
            dtype=h5py.special_dtype(vlen=str),
        )
        hf["features"][:] = list_training_features
        logging_utils.print_green("Saved features:", ",".join(list_training_features))

        # Save training features to hdf5
        logging_utils.print_green("Save data features to HDF5")
        arr_feat = df[list_training_features].values
        hf["data"].attrs["n_features"] = len(list_training_features)
        for idx, idx_pair in enumerate(
            tqdm(list_start_end, desc="Filling hdf5", ncols=100)
        ):
            arr = arr_feat[idx_pair[0] : idx_pair[1]]
            hf["data"][idx] = np.ravel(arr)

        # save data types for training
        try:
            hf["data_types_training"] = np.asarray(settings.data_types_training).astype(
                np.dtype("S100")
            )
        except Exception:
            hf["data_types_training"] = f"{settings.data_types_training}"
